<?php

/**
 * Copyright (C) 2018 Swashata Ghosh <swashata@wpquark.com>
 *
 * This file is part of eForm - WordPress Builder.
 *
 * eForm - WordPress Builder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * eForm - WordPress Builder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with eForm - WordPress Builder.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @package EForm
 * @subpackage Handler
 */
namespace WPEForm\Handler;

use  WPEForm\Auth\IAuth ;
use  WPEForm\Model\Form ;
use  WPEForm\Exception\ResourceDoesNotExistException ;
use function  WPEForm\Helpers\break_strings_with ;
use function  WPEForm\Helpers\convert_dictionary_to_list ;
use function  WPEForm\Helpers\convert_list_to_dictionary ;
use function  WPEForm\Helpers\get_element_from_array_by_id ;
use  WPEForm\Factory\Form\Elements ;
use  LogicException ;
use function  WPEForm\Helpers\extract_ids ;
use function  WPEForm\Helpers\format_datetime ;
use function  WPEForm\Helpers\format_time_from_seconds ;
use function  WPEForm\Helpers\parse_args ;
use  WPEForm\Factory\Form\Config\Settings ;
use  WPEForm\Form\Config\Settings\ISettings ;
use  WPEForm\Factory\Form\Config\Integrations ;
use  WPEForm\Factory\Form\Config\Permissions ;
use  WPEForm\Form\Config\Integrations\IIntegrations ;
use  WPEForm\Auth\Form as FormAuth ;
use  WPEForm\Exception\SubmissionValidationException ;
use  WPEForm\Form\Config\Permissions\IPermissions ;
use  WPEForm\Util\SlateJS ;
use  WPEForm\Model\Submission as SubmissionModel ;
use  WPEForm\Auth\Submission as SubmissionAuth ;
use  WPEForm\Exception\InvalidOperationException ;
use  WPEForm\System\SystemEndpoint ;
use  WPEForm\Util\Mailer ;
use  WPEForm\Util\Pdf ;
use  WPEForm\Util\Style ;
use  WPEForm\View\Front\Form as FrontForm ;
use  Exception ;
use  WPEForm\System\Init ;
use  WPEForm\Util\HTMLChart ;
// @codeCoverageIgnoreStart
if ( !defined( 'ABSPATH' ) ) {
    die( '' );
}
// @codeCoverageIgnoreEnd
/**
 * A class to handle all elements, integrations and payments of a submission.
 * Internally it calls proper controller and model to handle form data.
 */
class Submission
{
    /**
     * Get submission handler instance from submission model, auth model and update type.
     *
     * @param SubmissionModel $model Submission model, where id/token is already set.
     * @param SubmissionAuth  $auth Submission Auth instance.
     * @param bool            $is_update Whether or not this is an update.
     * @param bool            $admin_update Whether or not this admin update.
     * @param Form            $form_model A form model instance with data set. If not provided, one will be created.
     * @return null|Submission Handler instance on success, null on failure.
     */
    public static function get_instance_from_models(
        SubmissionModel $model,
        SubmissionAuth $auth,
        bool $is_update,
        bool $admin_update,
        ?Form $form_model = null
    ) : ?Submission
    {
        try {
            // bail early if submission model doesn't exist
            if ( !$model->exists() ) {
                return null;
            }
            // get data coz we need the formId from the submission model
            $submission_data = $model->read();
            $handler = new Submission(
                (int) $submission_data['formId'],
                $submission_data['elements'],
                true,
                $auth,
                $form_model
            );
            // set the token and other values
            $handler->set_token( $submission_data['token'] );
            $handler->set_referer( $submission_data['referer'] ?? '' );
            $handler->set_date( $submission_data['created'] ?? '0000-00-00 00:00:00' );
            $handler->set_submission_id( (string) $model->get_id() );
            $handler->set_time( $submission_data['time'] ?? 0 );
            $handler->set_admin_remarks( $submission_data['adminRemarks'] );
            $handler->set_submission_model( $model );
            $paid = \intval( $submission_data['paid'] );
            $handler->set_paid( ( \in_array( $paid, [ 0, 1, 2 ] ) ? $paid : 0 ) );
            
            if ( !\is_null( $submission_data['maxScore'] ) ) {
                $handler->set_score_from_db(
                    true,
                    \floatval( $submission_data['score'] ),
                    \floatval( $submission_data['maxScore'] ),
                    $submission_data['scoreData'] ?? []
                );
            } else {
                $handler->set_score_from_db(
                    false,
                    0,
                    0,
                    []
                );
            }
            
            // run the mutations because we need the configurations and primitives
            // of the submission handler set.
            $handler->run_settings_mutations( $is_update, $admin_update );
            $handler->run_permissions_presave_mutations( $is_update );
            return $handler;
        } catch ( Exception $e ) {
            // do nothing
        }
        return null;
    }
    
    /**
     * Get a Submission handler instance from token. It would return null if
     * token is invalid.
     *
     * @param string $token The token.
     * @param bool   $is_update Whether or not this handler is intended for update.
     * @param bool   $admin_update Whether or not the update is from admin.
     *
     * @return null|Submission Submission instance, or null if token is invalid.
     */
    public static function get_instance_from_token( string $token, bool $is_update, bool $admin_update ) : ?Submission
    {
        try {
            $model = new SubmissionModel();
            $auth = new SubmissionAuth();
            // set the token in the model
            $model->set_id( null, $token );
            return self::get_instance_from_models(
                $model,
                $auth,
                $is_update,
                $admin_update
            );
        } catch ( Exception $e ) {
            // do nothing, because the token is wrong, so just continue and
            // return null
        }
        return null;
    }
    
    /**
     * Get a Submission handler instance from token. It would return null if
     * token is invalid.
     *
     * @param string $id The token.
     * @param bool   $is_update Whether or not this handler is intended for update.
     * @param bool   $admin_update Whether or not the update is from admin.
     *
     * @return null|Submission Submission instance, or null if token is invalid.
     */
    public static function get_instance_from_id( string $id, bool $is_update, bool $admin_update ) : ?Submission
    {
        try {
            $model = new SubmissionModel();
            $auth = new SubmissionAuth();
            // set the token in the model
            $model->set_id( $id );
            return self::get_instance_from_models(
                $model,
                $auth,
                $is_update,
                $admin_update
            );
        } catch ( Exception $e ) {
            // do nothing, because the token is wrong, so just continue and
            // return null
        }
        return null;
    }
    
    /**
     * Form model for the submission handler.
     *
     * @var Form
     */
    protected  $form_model = null ;
    /**
     * Submission model. This is set when called using get_instance_from_model.
     *
     * @var SubmissionModel | null
     */
    protected  $submission_model = null ;
    /**
     * Auth for Submission handler.
     *
     * @var IAuth
     */
    protected  $auth = null ;
    /**
     * Submission element data for this submission handler from the user input.
     * This is not supposed to be processed.
     *
     * @var array
     */
    protected  $raw_elements_data = null ;
    /**
     * Submission element data for this handler which is processed by the class.
     *
     * @var array
     */
    protected  $processed_elements_data = null ;
    /**
     * Processed elements dictionary calculated from processed_elements_data.
     *
     * @var array
     */
    protected  $processed_elements_dictionary = [] ;
    /**
     * Elements dictionary as used from Form Model.
     *
     * @var array
     */
    protected  $elements_config_dictionary = [] ;
    /**
     * Score obtained throughout all elements for the primary score.
     *
     * @var float
     */
    protected  $score_obtained = 0 ;
    /**
     * Max possible score for the primary score.
     *
     * @var float
     */
    protected  $max_score = 0 ;
    /**
     * Whether or not this element has score.
     *
     * @var bool
     */
    protected  $has_score = false ;
    /**
     * Score data which stores all possible scores.
     *
     * @var array
     */
    protected  $score_data = [] ;
    /**
     * Primary score outcome name. Should be set by the mutators.
     *
     * @var array|null
     */
    protected  $score_metadata = null ;
    /**
     * Token for the submission. This must be set by the controller.
     *
     * @var string
     */
    protected  $token = null ;
    /**
     * ID for the submission. This must be set by the controller.
     *
     * @var string
     */
    protected  $submission_id = null ;
    /**
     * A variable to hold children of all conditionally hidden parents. We use
     * this to bypass and reset values of all children of elements which are
     * conditionally hidden.
     *
     * @var array
     */
    protected  $conditionally_hidden_by_parent = [] ;
    /**
     * Referer value.
     *
     * @var string|null
     */
    protected  $referer = null ;
    /**
     * Date value.
     *
     * @var string|null
     */
    protected  $date = null ;
    /**
     * Mentions for score designation and mentions.
     *
     * @var array
     */
    protected  $score_designation_mentions = [] ;
    /**
     * Submission time value.
     *
     * @var int
     */
    protected  $time = 0 ;
    /**
     * Whether or not the submission was made with a timer.
     *
     * @var bool
     */
    protected  $timer = false ;
    /**
     * Whether or not can user edit this submission (for existing submissions only).
     *
     * @var bool
     */
    protected  $can_user_edit = false ;
    /**
     * Administrator remarks.
     *
     * @var string
     */
    protected  $admin_remarks = '' ;
    /**
     * Score settings.
     *
     * Some settings class should override this.
     *
     * @var array
     */
    protected  $score_settings = [] ;
    /**
     * Set admin remarks.
     *
     * @param string $remarks Remarks as written by admin.
     * @return void
     */
    public function set_admin_remarks( string $remarks ) : void
    {
        $this->admin_remarks = $remarks;
    }
    
    /**
     * Get admin remarks. Could be an empty string.
     *
     * @return string
     */
    public function get_admin_remarks() : string
    {
        return $this->admin_remarks;
    }
    
    /**
     * Set whether or not can user edit.
     *
     * @param bool $can Flag to determine whether can user edit this submission.
     * @return void
     */
    public function set_can_user_edit( bool $can ) : void
    {
        $this->can_user_edit = $can;
    }
    
    /**
     * Get whether can user edit this submission or not.
     *
     * @return bool
     */
    public function get_can_user_edit() : bool
    {
        return $this->can_user_edit;
    }
    
    /**
     * Set the timer ON or OFF.
     *
     * @param bool $timer Timer, ON if passed true.
     * @return void
     */
    public function set_timer( bool $timer ) : void
    {
        $this->timer = $timer;
    }
    
    /**
     * Get the timer. True if it is ON, False other otherwise.
     *
     * @return bool
     */
    public function get_timer() : bool
    {
        return $this->timer;
    }
    
    /**
     * Set submission time.
     *
     * @param int|null $time The recorded time.
     * @return void
     */
    public function set_time( int $time ) : void
    {
        // convert to absolute integer
        $time = \absint( $time );
        $this->time = $time;
    }
    
    /**
     * Get submission time.
     *
     * @return null|int Submission time.
     */
    public function get_time() : int
    {
        return $this->time;
    }
    
    /**
     * Set score and designation mentions data.
     *
     * @param array $mentions The mentions data.
     *
     * @return void
     */
    public function set_score_and_designation_mentions( array $mentions ) : void
    {
        $this->score_designation_mentions = $mentions;
    }
    
    /**
     * Get score and designation mentions.
     *
     * @return array
     */
    public function get_score_and_designation_mentions() : array
    {
        return $this->score_designation_mentions;
    }
    
    /**
     * Set score settings.
     *
     * @param array $settings Score settings as per config. Check FormSettingsScoreInput.
     * @return void
     */
    public function set_score_settings( array $settings ) : void
    {
        $this->score_settings = $settings;
    }
    
    /**
     * Get score settings.
     *
     * @return array
     */
    public function get_score_settings() : array
    {
        return $this->score_settings;
    }
    
    /**
     * Set primary score outcome name.
     *
     * @param array{label: string, obtained: string, outof: string} $data Score data.
     * @return void
     */
    public function set_primaryscore_metadata( array $data ) : void
    {
        $this->score_metadata = $data;
    }
    
    /**
     * Get the referer.
     *
     * @throws LogicException If referer has not already been set.
     *
     * @return string
     */
    public function get_referer() : string
    {
        if ( !$this->referer ) {
            throw new LogicException( 'Referer is not set.' );
        }
        return $this->referer;
    }
    
    /**
     * Set the referer.
     *
     * @throws LogicException If the referer is already set.
     *
     * @param string $referer The referer.
     *
     * @return void
     */
    public function set_referer( string $referer )
    {
        if ( $this->referer ) {
            throw new LogicException( 'Referer is already set.' );
        }
        $this->referer = $referer;
    }
    
    /**
     * Get the date.
     *
     * @throws LogicException If date has not already been set.
     *
     * @return string
     */
    public function get_date() : string
    {
        if ( !$this->date ) {
            throw new LogicException( 'Date is not set.' );
        }
        return $this->date;
    }
    
    /**
     * Set the date.
     *
     * @throws LogicException If the date is already set.
     *
     * @param string $date The date.
     *
     * @return void
     */
    public function set_date( string $date )
    {
        if ( $this->date ) {
            throw new LogicException( 'Date is already set.' );
        }
        $this->date = $date;
    }
    
    /**
     * Primitives of the submission. They usually correspond to some direct db
     * table column.
     *
     * @var array
     */
    protected  $primitives = [
        'fName'  => null,
        'lName'  => null,
        'email'  => null,
        'phone'  => null,
        'ip'     => null,
        'userId' => null,
    ] ;
    /**
     * User email config. Depending on this, we handle the output of
     * `get_user_email`. One of the settings class must override this on
     * first pass.
     *
     * @var array
     */
    protected  $user_email_config = [
        'header'         => '',
        'subject'        => '',
        'message'        => '',
        'showSubmission' => true,
        'viewOnline'     => true,
        'viewOnlineText' => 'View Online',
        'footer'         => 'You are receiving this email because you have submitted a form.',
        'style_id'       => 'vividblue',
    ] ;
    /**
     * Configuration for administrator email. Depending on this, we handle the
     * output of `get_admin_email_mailer`. One of the settings class must override
     * this on run_mutation.
     *
     * @var (string|true)[]
     */
    protected  $admin_email_config = [
        'header'         => '',
        'subject'        => '',
        'message'        => '',
        'showSubmission' => true,
        'footer'         => 'You are receiving this email because you have submitted a form.',
        'style_id'       => 'vividblue',
    ] ;
    /**
     * Configure user email variables. Should be called during `run_mutation` and
     * not after.
     *
     * @param array $config User email config.
     * @return void
     */
    public function configure_user_email( array $config ) : void
    {
        $this->user_email_config = parse_args( $config, $this->user_email_config );
    }
    
    /**
     * Configure admin email variables. Should be called during `run_mutation`.
     *
     * @param array $config Admin user email config.
     * @return void
     */
    public function configure_admin_email( array $config ) : void
    {
        $this->admin_email_config = parse_args( $config, $this->admin_email_config );
    }
    
    /**
     * Get submission link of the submission.
     *
     * @return string Link of the submission.
     * @throws LogicException If token is not set.
     */
    public function get_submission_link() : string
    {
        return SystemEndpoint::get_summary_page_link( $this->get_token() );
    }
    
    /**
     * Get submission download link of the submission.
     *
     * @param bool $as_admin Whether the download link should be for admin.
     * @return string Link of the submission pdf.
     * @throws LogicException If token is not set.
     */
    public function get_submission_download_link( bool $as_admin ) : ?string
    {
        // ENDFS: this if block will be removed
        return null;
    }
    
    /**
     * Get submission print page link of the submission.
     *
     * @param bool $as_admin Whether the download link should be for admin.
     * @return string Link of the submission HTML.
     * @throws LogicException If token is not set.
     */
    public function get_submission_print_link( bool $as_admin ) : string
    {
        return SystemEndpoint::get_summary_print_page_link( ( $as_admin ? $this->get_submission_id() : $this->get_token() ), $as_admin );
    }
    
    /**
     * Get email attributes.
     *
     * @param bool $remarks Whether or not to include remarks. Default false.
     * @return array Array of attributes, to be passed directly to Mailer.
     * @throws LogicException If token is not set.
     */
    protected function get_email_attributes( $remarks = false ) : array
    {
        $submission_link = $this->get_submission_link();
        $form = $this->get_form_model();
        $attrs = [
            [
            'label' => \__( 'Form', 'wp-eform' ),
            'value' => \esc_html( $form->read()['name'] ),
        ],
            [
            'label' => \__( 'Submission Id', 'wp-eform' ),
            'value' => '<code>' . $this->get_token() . '</code>',
        ],
            [
            'label' => \__( 'Submission Date', 'wp-eform' ),
            'value' => format_datetime( \strtotime( $this->get_date() ) ),
        ],
            [
            'label' => \__( 'Record', 'wp-eform' ),
            'value' => '<a class="url" href="' . \esc_attr( $submission_link ) . '" target="_blank">' . break_strings_with( $submission_link, 37 ) . '</a>',
        ]
        ];
        
        if ( $remarks ) {
            $admin_remarks = $this->get_admin_remarks();
            if ( !empty($admin_remarks) ) {
                $attrs[] = [
                    'label' => \__( 'Remarks', 'wp-eform' ),
                    'value' => \esc_html( $admin_remarks ),
                ];
            }
        }
        
        if ( $this->has_score && $this->score_metadata ) {
            \array_splice(
                $attrs,
                2,
                0,
                [ [
                'label' => $this->score_metadata['label'],
                'value' => $this->score_metadata['obtained'] . \_x( ' / ', 'score separator', 'wp-eform' ) . $this->score_metadata['outof'],
            ] ]
            );
        }
        // If time is recorded
        if ( $this->time ) {
            $attrs[] = [
                'label' => \__( 'Duration', 'wp-eform' ),
                'value' => format_time_from_seconds( $this->time ),
            ];
        }
        return $attrs;
    }
    
    /**
     * Get a PDF instance to render the summary of the submission.
     *
     * @param bool $is_admin Whether or not the request is made by an admin.
     * @return Pdf
     */
    public function get_summary_pdf( bool $is_admin ) : Pdf
    {
        $form_data = $this->form_model->read();
        $style = new Style( $this->user_email_config['style_id'] );
        $submission_link = $this->get_submission_link();
        // attributes
        $attrs = $this->get_email_attributes( true );
        $title = $form_data['name'] . '<span> / ' . \get_bloginfo( 'name' ) . '</span>';
        $time_zone = \wp_timezone()->getName();
        $pdf = new Pdf();
        $pdf->set_style( $style )->page_title( $form_data['name'] )->title( $title )->body( $this->get_submission_preview_html( 'pdf' ) )->footer( \sprintf(
            /* translators: %1$s is time, %2$s is submission link */
            \__( 'generated on %1$s from<br />%2$s', 'wp-eform' ),
            format_datetime( \time(), true ) . (( !empty($time_zone) ? ' (' . $time_zone . ')' : '' )),
            '<a href="' . \esc_attr( $submission_link ) . '">' . break_strings_with( $submission_link, 58 ) . '</a>'
        ) )->attrs( $attrs );
        return $pdf;
    }
    
    /**
     * Get a safe PDF name.
     *
     * @return string
     */
    public function get_pdf_name() : string
    {
        $date = \gmdate( 'y-m-d', \strtotime( $this->get_date() ) );
        return 'submission-' . $date . '-' . \mb_strimwidth( $this->get_token(), 0, 8 ) . '.pdf';
    }
    
    /**
     * Get Mailer instance for sending out or dealing with User Notification
     * Email.
     *
     * @param bool $is_update Whether or not for the update.
     * @param bool $admin_update       Whether or not this is an admin update.
     *
     * @return Mailer Mailer instance with configured values.
     */
    public function get_user_email_mailer( bool $is_update, bool $admin_update ) : Mailer
    {
        $form_data = $this->form_model->read();
        // get all slatejs related stuff
        $header = $this->get_slatejs_html( $this->user_email_config['header'] );
        $footer = $this->get_slatejs_html( $this->user_email_config['footer'] );
        $subject = $this->get_slatejs_text( $this->user_email_config['subject'] );
        $message = $this->get_slatejs_html( $this->user_email_config['message'] );
        // other stuff
        $style = new Style( $this->user_email_config['style_id'] );
        $referer = $this->get_referer();
        $submission_link = $this->get_submission_link();
        // attributes
        $attrs = $this->get_email_attributes();
        // create mailer and set data
        $mailer = new Mailer();
        $mailer->set_style( $style )->body( $message )->heading( \esc_html( $form_data['name'] ), $referer )->subject( $subject )->header( $header )->footer( $footer )->email_url( SystemEndpoint::get_user_email_link( $this->get_token(), $is_update, $admin_update ) )->attrs( $attrs );
        if ( $this->user_email_config['viewOnline'] ) {
            $mailer->cta( $this->user_email_config['viewOnlineText'], $submission_link );
        }
        // get preview HTML is needed
        if ( $this->user_email_config['showSubmission'] ) {
            $mailer->data( $this->get_submission_preview_html( 'email' ) );
        }
        return $mailer;
    }
    
    /**
     * Get Mailer instance for sending out or dealing with Admin Notification
     * Email.
     *
     * @param bool $is_update Whether or not for the update.
     *
     * @return Mailer Mailer instance with configured values.
     */
    public function get_admin_email_mailer( bool $is_update ) : Mailer
    {
        $form_data = $this->form_model->read();
        // get all slatejs related stuff
        $header = $this->get_slatejs_html( $this->admin_email_config['header'], true );
        $footer = $this->get_slatejs_html( $this->admin_email_config['footer'], true );
        $subject = $this->get_slatejs_text( $this->admin_email_config['subject'], true );
        $message = $this->get_slatejs_html( $this->admin_email_config['message'], true );
        // other stuff
        $style = new Style( $this->admin_email_config['style_id'] );
        $referer = $this->get_referer();
        $submission_admin_link = SystemEndpoint::get_submission_admin_link( (string) $this->get_submission_id() );
        // attributes
        $attrs = $this->get_email_attributes();
        // create mailer and set data
        $mailer = new Mailer();
        $mailer->set_style( $style )->body( $message )->heading( \esc_html( $form_data['name'] ), $referer )->subject( $subject )->header( $header )->footer( $footer )->email_url( SystemEndpoint::get_admin_email_link( $this->get_token(), $is_update ) )->attrs( $attrs )->cta( \__( 'View Submission', 'wp-eform' ), $submission_admin_link );
        // get preview HTML is needed
        if ( $this->admin_email_config['showSubmission'] ) {
            $mailer->data( $this->get_submission_preview_html( 'email' ) );
        }
        return $mailer;
    }
    
    /**
     * Payment status flag of the submission.
     *
     * 0 - Unpaid.
     * 1 - Paid.
     * 2 - Not needed.
     *
     * @var int
     */
    protected  $paid = 2 ;
    /**
     * Throw if the passed in `$primitive` isn't a valid one.
     *
     * @throws LogicException If passed in `$primitive` isn't valid.
     *
     * @param string $primitive The primitive.
     *
     * @return void
     */
    protected function throw_if_not_valid_primitive( string $primitive ) : void
    {
        if ( !\array_key_exists( $primitive, $this->primitives ) ) {
            throw new LogicException( "{$primitive} is not a valid primitive." );
        }
    }
    
    /**
     * Get the token.
     *
     * @throws LogicException If token has not already been set.
     *
     * @return string
     */
    public function get_token() : string
    {
        if ( !$this->token ) {
            throw new LogicException( 'Token is not set.' );
        }
        return $this->token;
    }
    
    /**
     * Set the token.
     *
     * @throws LogicException If the token is already set.
     *
     * @param string $token The token.
     *
     * @return void
     */
    public function set_token( string $token )
    {
        if ( $this->token ) {
            throw new LogicException( 'Token is already set.' );
        }
        $this->token = $token;
    }
    
    /**
     * Check if token has been set or not.
     *
     * @return bool
     */
    public function has_token() : bool
    {
        if ( $this->token ) {
            return true;
        }
        return false;
    }
    
    /**
     * Get the submission_id.
     *
     * @throws LogicException If submission_id has not already been set.
     *
     * @return string
     */
    public function get_submission_id() : string
    {
        if ( !$this->submission_id ) {
            throw new LogicException( 'submission_id is not set.' );
        }
        return $this->submission_id;
    }
    
    /**
     * Set the submission_id.
     *
     * @throws LogicException If the submission_id is already set.
     *
     * @param string $submission_id The submission_id.
     *
     * @return void
     */
    public function set_submission_id( string $submission_id )
    {
        if ( $this->submission_id ) {
            throw new LogicException( 'submission_id is already set.' );
        }
        $this->submission_id = $submission_id;
    }
    
    /**
     * Check if submission_id has been set or not.
     *
     * @return bool
     */
    public function has_submission_id() : bool
    {
        if ( $this->submission_id ) {
            return true;
        }
        return false;
    }
    
    /**
     * Get a primitive value from the handler.
     *
     * @throws LogicException If passed in `$primitive` isn't valid.
     *
     * @param string $primitive The primitive, `fName`, `lName`, `email`, `phone`,
     *                          `ip` or `userId`.
     *
     * @return string|null
     */
    public function get_primitive( string $primitive ) : ?string
    {
        $this->throw_if_not_valid_primitive( $primitive );
        return $this->primitives[$primitive];
    }
    
    /**
     * Get a primitive value on the handler.
     *
     * @throws LogicException If passed in `$primitive` isn't valid.
     *
     * @param string      $primitive The primitive, `fName`, `lName`, `email`, `phone`
     *                               or `ip`.
     * @param string|null $value     The value of the primitive.
     *
     * @return void
     */
    public function set_primitive( string $primitive, ?string $value ) : void
    {
        $this->throw_if_not_valid_primitive( $primitive );
        $this->primitives[$primitive] = $value;
    }
    
    /**
     * Get the callback for SlateJS mentions renderer scoped to the form and
     * submission.
     *
     * @return callable Callback.
     */
    public function get_slatejs_mentions_callback() : callable
    {
        return [ $this, 'slatejs_mentions_renderer' ];
    }
    
    /**
     * Get element simple value by element Id.
     *
     * @param string $element_id The Id of the element.
     * @param bool   $nested Whether this is nested. In that case, skip futher mentions render.
     * @return string Simple Value.
     */
    public function get_element_simple_value( string $element_id, bool $nested = false ) : string
    {
        // Now we might have an Element, so we need to extract value
        $element_data = $this->elements_config_dictionary[$element_id] ?? null;
        // If no element data found, then bail
        if ( !$element_data ) {
            return '';
        }
        // get the submission data
        $element_submission_data = $this->processed_elements_dictionary[$element_id] ?? null;
        // If no element submission data, then bail
        if ( !$element_submission_data ) {
            return '';
        }
        // if conditionally hidden, then bail
        if ( $element_submission_data['conditionallyHidden'] ) {
            return '';
        }
        $element_type = $element_data['type'];
        if ( !Elements::has_element( $element_type ) ) {
            // bail
            return '';
        }
        $element_instance = Elements::get_element_instance( $element_type );
        return $element_instance->get_simple_value(
            $element_data['config'][$element_type],
            $element_submission_data['value'][$element_type],
            [
            'type'   => 'label',
            'return' => 'string',
        ],
            ( $nested ? null : $this->get_slatejs_mentions_callback() )
        );
    }
    
    /**
     * A mentions renderer scoped to the form and submission data. This can be
     * used directly from the instance to a SlateJS model.
     *
     * It is upto the caller to properly set the `$nested` argument to prevent any
     * circular calls.
     *
     * @param string $element_id ID of the element/mention.
     * @param bool   $nested If true, then it will not render another element.
     *
     * @return string Rendered value.
     *
     * @throws LogicException If submission Id is not set.
     */
    public function slatejs_mentions_renderer( string $element_id, bool $nested = false ) : string
    {
        // first take care of the built-in ones
        $form_data = $this->form_model->read();
        $first_name = $this->get_primitive( 'fName' ) ?? \__( 'Anonymous', 'wp-eform' );
        $last_name = $this->get_primitive( 'lName' );
        $full_name = $first_name . (( $last_name ? ' ' . $last_name : '' ));
        switch ( $element_id ) {
            case '::submissionId::':
                return $this->get_submission_id();
            case '::formId::':
                return (string) $this->form_model->get_id();
            case '::formName::':
                return $form_data['name'];
            case '::invoiceId::':
                return '';
            case '::paymentFrequency::':
                return '';
            case '::paymentInterval::':
                return '';
            case '::paymentGatewayName::':
                return '';
            case '::paymentInvoice::':
                return '';
            case '::paymentStatus::':
                $payment_status = $this->get_paid();
                return ( $payment_status === 0 ? \__( 'unpaid', 'wp-eform' ) : (( $payment_status === 1 ? \__( 'paid', 'wp-eform' ) : \__( 'not needed', 'wp-eform' ) )) );
            case '::siteName::':
                return \get_bloginfo( 'name' );
            case '::name::':
                return $full_name;
            case '::fName::':
                return $first_name;
            case '::lName::':
                return $last_name;
            case '::email::':
                return $this->get_primitive( 'email' ) ?? '';
            case '::phone::':
                return $this->get_primitive( 'phone' ) ?? '';
            case '::totalSubmissions::':
                return Form::get_total_submission_count( $this->form_model->get_id() );
            case '::adminRemarks::':
                return \esc_html( $this->get_admin_remarks() );
        }
        // process the score related stuff
        if ( isset( $this->score_designation_mentions[$element_id] ) ) {
            return $this->score_designation_mentions[$element_id];
        }
        return $this->get_element_simple_value( $element_id, $nested );
    }
    
    /**
     * Get slatejs dynamic anchor tags.
     *
     * @param bool $is_admin If true, it will also add admin specific links.
     * @return array Dyamic anchor tags.
     */
    public function get_slatejs_dynamic_anchor_tags( bool $is_admin ) : array
    {
        $form_id = (int) $this->form_model->get_id();
        $tags = [
            '%SUBMISSION_LINK%' => $this->get_submission_link(),
            '%FORM_LINK%'       => $this->referer ?? FrontForm::get_form_standalone_link( $form_id ),
        ];
        
        if ( $is_admin ) {
            $tags['%FORM_ADMIN_LINK%'] = SystemEndpoint::get_standalone_admin_link( '/forms/edit/' . (string) $form_id );
            $tags['%ADMIN_SUBMISSION_LINK%'] = SystemEndpoint::get_submission_admin_link( (string) $this->get_submission_id() );
        }
        
        return $tags;
    }
    
    /**
     * Get Text from a SlateJS editor children json (as stored in model.)
     *
     * @param string     $editor_json Editor JSON.
     * @param bool       $is_admin Whether or not for administrator.
     * @param string     $mode Render mode.
     * @param null|array $exclude_mentions_characters Mentions characters to exlcude.
     * @param bool       $render_mentions Whether to disable rendering mentions.
     * @return string Rendered Text.
     */
    protected function get_slatejs_text(
        string $editor_json,
        bool $is_admin = false,
        string $mode = 'singleline',
        ?array $exclude_mentions_characters = null,
        bool $render_mentions = true
    ) : string
    {
        return SlateJS::parseEditorChildrenJSON( $editor_json )->setRenderMentions( $this->get_slatejs_mentions_callback() )->setDynamicAnchorTags( $this->get_slatejs_dynamic_anchor_tags( $is_admin ) )->getText( $mode, $exclude_mentions_characters, $render_mentions );
    }
    
    /**
     * Get HTML from a SlateJS editor children json (as stored in model.)
     *
     * @param string     $editor_json Editor JSON.
     * @param bool       $is_admin Whether or not for administrator.
     * @param string     $mode Render mode.
     * @param null|array $exclude_mentions_characters Mentions characters to exlcude.
     * @param bool       $render_mentions Whether to disable rendering mentions.
     * @return string Rendered HTML.
     */
    protected function get_slatejs_html(
        string $editor_json,
        bool $is_admin = false,
        string $mode = 'multiline',
        ?array $exclude_mentions_characters = null,
        bool $render_mentions = true
    ) : string
    {
        return SlateJS::parseEditorChildrenJSON( $editor_json )->setRenderMentions( $this->get_slatejs_mentions_callback() )->setDynamicAnchorTags( $this->get_slatejs_dynamic_anchor_tags( $is_admin ) )->getHTML( $mode, $exclude_mentions_characters, $render_mentions );
    }
    
    /**
     * Recursively extract elements, if
     * 1. Element is not conditionally hidden.
     * 2. Element is not a design/security element.
     *
     * @param string $element_id Element Id to extract.
     * @return array List of elements, in order.
     */
    protected function recursively_extract_active_submission_elements( string $element_id ) : array
    {
        $elements = [];
        // get config and submission from lookup
        $element_config = $this->elements_config_dictionary[$element_id] ?? null;
        $element_submission = $this->processed_elements_dictionary[$element_id] ?? null;
        // if element is not active, then bail
        if ( !$element_config || !$element_submission ) {
            return $elements;
        }
        // if element is conditionally hidden, then bail (also all child)
        if ( $element_submission['conditionallyHidden'] === true ) {
            return $elements;
        }
        // check the element type
        $element_type = $element_config['type'];
        if ( !Elements::has_element( $element_type ) ) {
            return $elements;
        }
        $element_instance = Elements::get_element_instance( $element_type );
        // Now add this element, if it is of type choice or input
        if ( \in_array( $element_instance->get_definition()['category'], [ 'choice', 'input' ] ) ) {
            $elements[] = $element_id;
        }
        // Now if the element has children, then process them too
        if ( isset( $element_config['children'] ) && !empty($element_config['children']) ) {
            foreach ( $element_config['children'] as $child_element_id ) {
                $active_submission_elements = $this->recursively_extract_active_submission_elements( $child_element_id );
                // we need to check this because it will fail in PHP 7.1 and 7.2
                if ( \count( $active_submission_elements ) ) {
                    \array_push( $elements, ...$active_submission_elements );
                }
            }
        }
        return $elements;
    }
    
    /**
     * Get active structures of a form, given the form data and submission data.
     *
     * @return array Array of structures.
     */
    protected function get_active_structures() : array
    {
        $this->throw_if_elements_not_processed();
        $form_data = $this->form_model->read();
        $structures = $form_data['structures'];
        return $structures;
    }
    
    /**
     * Extract active elements from the form model and submission data for the
     * given stricture id.
     *
     * @param string $structure_id Id of the stricture.
     *
     * @return array Array of elements, in original order.
     */
    protected function get_active_submission_elements_per_structure( string $structure_id ) : array
    {
        $this->throw_if_elements_not_processed();
        $elements = [];
        $form_data = $this->form_model->read();
        $structures = $form_data['structures'];
        if ( empty($structures) ) {
            return $elements;
        }
        // Now get the structure
        $structure = null;
        foreach ( $structures as $st ) {
            
            if ( $st['id'] === $structure_id ) {
                $structure = $st;
                break;
            }
        
        }
        if ( empty($structure) ) {
            return $elements;
        }
        $children = $structure['children'];
        if ( empty($children) ) {
            return $elements;
        }
        foreach ( $children as $element_id ) {
            $active_submission_elements = $this->recursively_extract_active_submission_elements( $element_id );
            // we need to check this because it will fail in PHP 7.1 and 7.2
            if ( \count( $active_submission_elements ) ) {
                \array_push( $elements, ...$active_submission_elements );
            }
        }
        return $elements;
    }
    
    /**
     * Get Submission Preview HTML.
     *
     * @param string $purpose Purpose of the HTML. Could be 'email' or 'pdf'.
     * @return string The HTML.
     */
    public function get_submission_preview_html( string $purpose = 'email' ) : string
    {
        $output = '';
        // ENDFS: this if block will be removed
        $structures = $this->get_active_structures();
        foreach ( $structures as $structure ) {
            $elements = $this->get_active_submission_elements_per_structure( $structure['id'] );
            // If elements are empty, then continue
            if ( empty($elements) ) {
                continue;
            }
            $output .= '<h2 class="page-title">' . SlateJS::parseEditorChildrenJSON( $structure['config']['title'] )->setRenderMentions( $this->get_slatejs_mentions_callback() )->getText( 'singleline' ) . '</h2>';
            foreach ( $elements as $element_id ) {
                $element_config = $this->elements_config_dictionary[$element_id] ?? null;
                $element_submission = $this->processed_elements_dictionary[$element_id] ?? null;
                if ( !$element_config || !$element_submission ) {
                    continue;
                }
                $element_type = $element_config['type'];
                if ( !Elements::has_element( $element_type ) ) {
                    continue;
                }
                $element_instance = Elements::get_element_instance( $element_type );
                $element_output = $element_instance->get_html_value(
                    $element_config['appearance'],
                    $element_config['config'][$element_type],
                    $element_submission['value'][$element_type],
                    $this->get_slatejs_mentions_callback(),
                    $purpose
                );
                
                if ( !empty($element_output) ) {
                    $output .= $element_output;
                    // also add score if present
                    
                    if ( isset( $element_submission['scores'] ) && !empty($element_submission['scores']) && $this->score_metadata !== null ) {
                        $output .= '<table class="scores" width="100%" cellpadding="0" cellspacing="0" role="presentation">';
                        foreach ( $element_submission['scores'] as $elem_score ) {
                            
                            if ( $elem_score['has'] && $elem_score['outof'] !== 0 ) {
                                $sid = $elem_score['scoreId'];
                                $output .= '<tr><td class="scores__cell scores__cell--heading">' . ($this->score_designation_mentions["::score-name-{$sid}::"] ?? \__( 'Score', 'wp-eform' )) . '</td><td class="scores__cell scores__cell--value">' . \call_user_func( $this->score_metadata['formatter'], $elem_score['obtained'] ) . '</td><td class="scores__cell scores__cell--value">' . \call_user_func( $this->score_metadata['formatter'], $elem_score['outof'] ) . '</td><td class="scores__cell scores__cell--value">' . \call_user_func( $this->score_metadata['formatter'], $elem_score['obtained'] / $elem_score['outof'] * 100 ) . '%' . '</td></tr>';
                                if ( $this->score_designation_mentions["::score-description-{$sid}::"] ) {
                                    $output .= '<tr><td colspan="4" class="scores__cell">' . $this->score_designation_mentions["::score-description-{$sid}::"] . '</td></tr>';
                                }
                            }
                        
                        }
                        $output .= '</table>';
                    }
                    
                    $output .= '<hr class="divider" />';
                }
            
            }
        }
        return $output;
    }
    
    /**
     * Set paid status in the handler.
     *
     * @throws LogicException If invalid $status is passed.
     *
     * @param int $status Payment status, one of 0, 1 or 2.
     *
     * @return void
     */
    public function set_paid( int $status ) : void
    {
        if ( !\in_array( $status, [ 0, 1, 2 ], true ) ) {
            throw new LogicException( 'Payment status could be one of 0, 1 and 2.' );
        }
        $this->paid = $status;
    }
    
    /**
     * Get paid status of the handler.
     *
     * ```
     * 0 - Unpaid.
     * 1 - Paid.
     * 2 - Not needed.
     * ```
     *
     * @return int
     */
    public function get_paid() : int
    {
        return $this->paid;
    }
    
    /**
     * Get paid status for GraphQL.
     *
     * @return null|bool
     */
    public function get_paid_for_graphql() : ?bool
    {
        if ( $this->paid === 2 ) {
            return null;
        }
        if ( $this->paid === 1 ) {
            return true;
        }
        return false;
    }
    
    /**
     * Reset score data on the instance.
     *
     * @return void
     */
    protected function reset_score() : void
    {
        $this->score_obtained = 0;
        $this->max_score = 0;
        $this->has_score = false;
        $this->score_data = [];
    }
    
    /**
     * Get score obtained from processed elements from the instance.
     * It returns null if elements don't have any scorable config.
     *
     * @return float|null
     */
    public function get_score_obtained() : ?float
    {
        if ( !$this->has_score ) {
            return null;
        }
        return $this->score_obtained;
    }
    
    /**
     * Get maximum score one can obtain from processed elements from the instance.
     * It returns null if elements don't have any scorable config.
     *
     * @return float|null
     */
    public function get_max_score() : ?float
    {
        if ( !$this->has_score ) {
            return null;
        }
        return $this->max_score;
    }
    
    /**
     * Get cumulative score data for all listed scores in the form.
     *
     * @param bool $dictionary Whether or not to return in dictionary format.
     *
     * @return array|null
     */
    public function get_score_data( $dictionary = false ) : ?array
    {
        if ( !$this->has_score ) {
            return null;
        }
        if ( $dictionary ) {
            return $this->score_data;
        }
        return convert_dictionary_to_list( $this->score_data, 'scoreId' );
    }
    
    /**
     * Set score data from db stored values.
     *
     * @param bool  $has_score Whether or not submission has score value.
     * @param float $score Score obtained.
     * @param float $max_score Maximum score to be obtained.
     * @param array $score_data Cumulative score data.
     *
     * @return void
     */
    public function set_score_from_db(
        bool $has_score,
        float $score = 0,
        float $max_score = 0,
        array $score_data
    ) : void
    {
        
        if ( !$has_score ) {
            $this->has_score = false;
            $this->max_score = 0;
            $this->score_obtained = 0;
            $this->score_data = [];
        } else {
            $this->has_score = true;
            $this->score_obtained = $score;
            $this->max_score = $max_score;
            $this->score_data = convert_list_to_dictionary( $score_data, 'scoreId' );
        }
    
    }
    
    /**
     * Throw a logic exception if elements data isn't processed yet.
     *
     * @throws LogicException If data hasn't been processed yet.
     *
     * @return void
     */
    protected function throw_if_elements_not_processed() : void
    {
        if ( $this->processed_elements_data === null ) {
            throw new LogicException( 'Elements are not processed yet, please call process_elements.' );
        }
    }
    
    /**
     * Get processed elements data from the handler.
     *
     * @throws LogicException If data hasn't been processed yet.
     *
     * @return array
     */
    public function get_processed_elements() : array
    {
        $this->throw_if_elements_not_processed();
        return $this->processed_elements_data;
    }
    
    /**
     * Set processed elements. This method makes it possibles to config hooks to
     * mutate data before it is being saved in the database. No checks are
     * performed in the handler level. However, the controller will pass it to
     * the model and the model will normalize and validate the data before saving.
     * So make sure not to mess up with the data structure.
     *
     * @param array $elements Processed elements data.
     *
     * @return void
     */
    public function set_processed_elements( array $elements ) : void
    {
        $this->processed_elements_data = $elements;
        $this->processed_elements_dictionary = convert_list_to_dictionary( $elements );
    }
    
    /**
     * Create an instance.
     *
     * @throws ResourceDoesNotExistException If the form does not exist.
     *
     * @param int   $form_id       Id of the Form.
     * @param array $elements_data Submission element data as found from user input
     *                             or database.
     * @param bool  $processed     Whether the element data is processed or not.
     * @param IAuth $auth          Auth instance.
     * @param Form  $form_model    Form Model instance, if not provided one will be created.
     */
    public function __construct(
        int $form_id,
        array $elements_data,
        bool $processed,
        IAuth $auth,
        ?Form $form_model = null
    )
    {
        
        if ( !$form_model ) {
            $form_model = new Form();
            $form_model->set_id( $form_id );
        }
        
        if ( !$form_model->exists() ) {
            throw new ResourceDoesNotExistException( 'The form for the submission does not exist.' );
        }
        $this->form_model = $form_model;
        
        if ( !$processed ) {
            $this->raw_elements_data = $elements_data;
        } else {
            $this->processed_elements_data = $elements_data;
            $this->processed_elements_dictionary = convert_list_to_dictionary( $elements_data );
        }
        
        $this->auth = $auth;
        // set the elements_config_dictionary
        $this->elements_config_dictionary = convert_list_to_dictionary( $this->form_model->read()['elements'] );
    }
    
    /**
     * Get Auth Instance.
     *
     * @return IAuth
     */
    public function get_auth() : IAuth
    {
        return $this->auth;
    }
    
    /**
     * Get form data as stored in the database.
     *
     * @return array|null
     */
    public function get_form_data() : ?array
    {
        if ( $this->form_model !== null ) {
            return $this->form_model->read();
        }
        return null;
    }
    
    /**
     * Get Form Model from the handler.
     *
     * @throws LogicException If for some reason the model is null.
     *
     * @return Form The Form Model.
     */
    public function get_form_model() : Form
    {
        if ( $this->form_model === null ) {
            throw new LogicException( 'Form model is not present in submission handler.' );
        }
        return $this->form_model;
    }
    
    /**
     * Get Submission Model from the handler.
     *
     * @throws LogicException If for some reason the model is null.
     *
     * @return  SubmissionModel Submission Model.
     */
    public function get_submission_model() : SubmissionModel
    {
        if ( $this->submission_model === null ) {
            throw new LogicException( 'Submission model is not present in submission handler.' );
        }
        return $this->submission_model;
    }
    
    /**
     * Set Submission Model from the handler.
     *
     * @param SubmissionModel $submission The submission model.
     *
     * @return  void.
     */
    public function set_submission_model( SubmissionModel $submission ) : void
    {
        $this->submission_model = $submission;
    }
    
    /**
     * Recursively extract elements from a single element by looping over it's
     * children. It also adds the parent element to the list.
     *
     * It doesn't ignore any types of elements, including design, because they
     * can also have conditionallyHidden state.
     *
     * @param string $element_id Id of the current element.
     *
     * @return array
     */
    protected function recursively_extract_elements( string $element_id ) : array
    {
        $all_elements = [];
        // if the element is a design element, then don't add it to the list
        $active_element = $this->elements_config_dictionary[$element_id];
        // bail early, if this element is not present in the registry anymore
        if ( !Elements::has_element( $active_element['type'] ) ) {
            return $all_elements;
        }
        $all_elements[] = $element_id;
        // now loop over children, if present and add them too
        if ( !empty($active_element['children']) ) {
            foreach ( $active_element['children'] as $child_element_id ) {
                $all_elements = \array_merge( $all_elements, $this->recursively_extract_elements( $child_element_id ) );
            }
        }
        return $all_elements;
    }
    
    /**
     * Get flattened active elements from form data. It starts with structure
     * where elements are initiated and loops over all structure and all elements
     * to get all set of active elements.
     *
     * It doesn't ignore any types of elements, including design, because they
     * can also have conditionallyHidden state.
     *
     * @param array $form_data Form Data.
     *
     * @return array
     */
    protected function get_flattened_elements( array $form_data ) : array
    {
        $structures = $form_data['structures'];
        $flattened_elements = [];
        // foreach elements in structure, get the active elements
        foreach ( $structures as $structure ) {
            foreach ( $structure['children'] as $element_id ) {
                $flattened_elements = \array_merge( $flattened_elements, $this->recursively_extract_elements( $element_id ) );
            }
        }
        return $flattened_elements;
    }
    
    /**
     * Run settings pre processors.
     *
     * @param bool $is_update Whether or not it is an update.
     * @param bool $admin_update       Whether or not this is an admin update.
     * @return void
     */
    public function run_settings_preprocessors( bool $is_update, bool $admin_update )
    {
        $registered_settings = Settings::get_registered_configs();
        $form_data = $this->get_form_data();
        $form_settings = $form_data['settings'];
        // Loop over all registered Settings config
        /** @var ISettings $s_instance */
        foreach ( $registered_settings as $s_key => $s_instance ) {
            // call run side effect on each and every one of them
            $s_instance->run_preprocessor(
                $this,
                $form_settings[$s_key],
                $is_update,
                $admin_update
            );
        }
    }
    
    /**
     * Check for any pre-emptive limitations the form might have. It simply
     * throws a GraphQL compatible error to be shown to the user.
     *
     * @throws SubmissionValidationException If there is any pre-emptive limitations.
     *
     * @return void
     */
    public function check_preemptive_limitations()
    {
        // Check for pre-emptive limitations
        $pre_emptive_limitations = Permissions::get_preemptive_limitations( $this->get_form_model(), new FormAuth() );
        if ( !empty($pre_emptive_limitations) ) {
            throw new SubmissionValidationException( SubmissionValidationException::getSlateJSON(
                $pre_emptive_limitations[0]['id'],
                $pre_emptive_limitations[0]['message'],
                $pre_emptive_limitations[0]['mentions'],
                $pre_emptive_limitations[0]['dynamicAnchors']
            ) );
        }
    }
    
    /**
     * Check for any post submission limitations the form might have. It will
     * throw a GraphQL compatible error which should be picked up by the client.
     * See render method for SubmissionValidationException for more.
     *
     * @throws SubmissionValidationException If there is any post submission limitations.
     *
     * @return void
     */
    public function check_postsubmission_limitations()
    {
        $post_limitations = Permissions::get_postsubmission_limitations( $this->get_form_model(), new FormAuth(), $this );
        if ( !empty($post_limitations) ) {
            throw new SubmissionValidationException( SubmissionValidationException::getSlateJSON(
                $post_limitations[0]['id'],
                $post_limitations[0]['message'],
                $post_limitations[0]['mentions'],
                $post_limitations[0]['dynamicAnchors']
            ) );
        }
    }
    
    /**
     * Throw a SubmissionValidationException if form element is null or the type
     * isn't registered.
     *
     * @param array|null $form_element The form element.
     * @param array      $submission_element The submission element accompanied.
     *
     * @throws SubmissionValidationException If form element is null or non-existent.
     * @return void
     */
    protected function throw_if_form_element_is_non_existent( ?array $form_element, array $submission_element )
    {
        if ( $form_element === null || !Elements::has_element( $form_element['type'] ) ) {
            throw new SubmissionValidationException( SubmissionValidationException::getSlateJSON(
                'nonExistentElement',
                SlateJS::parseHTML( '<p>' . sprintf(
                /* translators: %1$s is element Id, handled client-side */
                __( 'Element <mentions character="%1$s"></mentions> is non existent.', 'wp-eform' ),
                $submission_element['id']
            ) . '</p>' )->getEditorChildrenJSON(),
                null,
                null
            ) );
        }
    }
    
    /**
     * Reset conditionally hidden by parent variable. Call this at the beginning
     * of process elements.
     *
     * @return void
     */
    protected function reset_conditionally_hidden_by_parent()
    {
        $this->conditionally_hidden_by_parent = [];
    }
    
    /**
     * Calculate conditionally hidden by parent variable pre-emptively.
     *
     * @param array $form_data Form data as read from model.
     *
     * @return void
     */
    protected function calculate_conditionally_hidden_by_parent( array $form_data )
    {
        $this->reset_conditionally_hidden_by_parent();
        foreach ( $this->raw_elements_data as $submission_element ) {
            $element_id = $submission_element['id'];
            $form_element = get_element_from_array_by_id( $form_data['elements'], $element_id );
            // If form element itself is null or non-existent, then throw
            $this->throw_if_form_element_is_non_existent( $form_element, $submission_element );
            // Now if submission element is conditionally hidden, then add it and all
            // it's children
            if ( $submission_element['conditionallyHidden'] ) {
                $this->conditionally_hidden_by_parent = \array_merge( $this->conditionally_hidden_by_parent, $this->recursively_extract_elements( $element_id ) );
            }
        }
    }
    
    /**
     * Check whether an element is conditionally hidden by parent. It assumes
     * that the conditionally_hidden_by_parent has been calculated.
     *
     * @param string $element_id Element Id.
     *
     * @return bool
     */
    protected function is_element_conditionally_hidden_by_parent( string $element_id ) : bool
    {
        return \in_array( $element_id, $this->conditionally_hidden_by_parent, true );
    }
    
    /**
     * Get an empty score data with proper structure.
     *
     * @param string $id ID of the score.
     * @return array Empty score data.
     */
    public static function get_empty_score_data( string $id ) : array
    {
        return [
            'scoreId'  => $id,
            'has'      => false,
            'outof'    => 0.0,
            'obtained' => 0.0,
        ];
    }
    
    /**
     * Process elements of a submission taking into consideration the form itself.
     *
     * @throws SubmissionValidationException If element is invalid or validation logic fails.
     *
     * @return void
     */
    public function process_elements() : void
    {
        $this->reset_score();
        $form_data = $this->form_model->read();
        $this->calculate_conditionally_hidden_by_parent( $form_data );
        // get all scores from config
        $all_scores = $form_data['settings']['score']['scoreLists'];
        // primary score Id is fixed
        $primary_score_id = \WPEForm\Form\Config\Settings\Score::PRIMARY_SCORE_ID;
        if ( !empty($all_scores) ) {
            // we assume that primary score id is in the list, but if it is not,
            // then the check later on, doesn't add to any thing.
            foreach ( $all_scores as $score_dt ) {
                $this->score_data[$score_dt['id']] = $this->get_empty_score_data( $score_dt['id'] );
            }
        }
        $processed_elements = [];
        $active_elements = $this->get_flattened_elements( $form_data );
        // 1. Loop through all elements and
        foreach ( $this->raw_elements_data as $submission_element ) {
            // 1.a Get the corresponding form element
            $form_element = get_element_from_array_by_id( $form_data['elements'], $submission_element['id'] );
            // 1.b Check if element is valid and present in the form
            $this->throw_if_form_element_is_non_existent( $form_element, $submission_element );
            // check if the element id is in active set of elements
            // now some element from active set could be absent because it
            // could be conditionally hidden, but an element in submission can not
            // be outside active elements
            if ( !\in_array( $submission_element['id'], $active_elements ) ) {
                throw new SubmissionValidationException( SubmissionValidationException::getSlateJSON(
                    'nonActiveElement',
                    SlateJS::parseHTML( '<p>' . sprintf(
                    /* translators: %1$s is element Id, handled client-side */
                    __( 'Element <mentions character="%1$s"></mentions> is not active.', 'wp-eform' ),
                    $submission_element['id']
                ) . '</p>' )->getEditorChildrenJSON(),
                    null,
                    null
                ) );
            }
            // initialize what the processed element would like
            $processed_element = [
                'id'                  => $submission_element['id'],
                'value'               => Elements::get_element_data_value_structure_with_default( $form_element['type'] ),
                'scores'              => [],
                'conditionallyHidden' => $submission_element['conditionallyHidden'],
            ];
            // check if the element is conditionally hidden. If so, then don't do
            // validation, reset score and value,
            
            if ( $submission_element['conditionallyHidden'] || $this->is_element_conditionally_hidden_by_parent( $submission_element['id'] ) ) {
                $processed_elements[] = $processed_element;
                // bail from this element
                continue;
            }
            
            // If the submission element doesn't really need to have a value
            
            if ( !\in_array( $form_element['type'], \array_keys( $processed_element['value'] ) ) ) {
                // Then jut store the empty value and proceed
                $processed_elements[] = $processed_element;
                continue;
            }
            
            // extract the Element Config and Submission Value
            // we have to extract them from the unified interface because the Element
            // class does not expect it with the following structure
            // $config = [
            // 'checkbox' => [...],
            // 'radio' => [...],
            // ];
            // etc. It expects the config and submission value directly as per with
            // config_vars and data_value_vars. How we have actually got it from
            // GraphQL and/or Model, is irrelevant to the Element class.
            $element_config = $form_element['config'][$form_element['type']];
            $submission_value = $submission_element['value'][$form_element['type']] ?? null;
            // if submission value is null, then it's an error, because element is
            // not conditionally hidden
            if ( $submission_value === null ) {
                throw new SubmissionValidationException( SubmissionValidationException::getSlateJSON(
                    'nullSubmissionValue',
                    SlateJS::parseHTML( '<p>' . sprintf(
                    /* translators: %1$s is element Id, handled client-side */
                    __( 'Value cannot be null for element <mentions character="%1$s"></mentions>.', 'wp-eform' ),
                    $submission_element['id']
                ) . '</p>' )->getEditorChildrenJSON(),
                    null,
                    null
                ) );
            }
            // get an element instance for further dealings
            $element = Elements::get_element_instance( $form_element['type'] );
            // Normalize the submission value
            $submission_value = $element->get_normalized_data_value( $submission_value );
            // 1.c Check with element class for validation
            $validation_status = $element->get_validation_status( $element_config, $submission_value );
            
            if ( !$validation_status['success'] ) {
                // if we have timer on, then we just ditch the value of invalid elements
                // and move on
                
                if ( $this->get_timer() ) {
                    $processed_elements[] = $processed_element;
                    continue;
                }
                
                throw new SubmissionValidationException( SubmissionValidationException::getSlateJSON(
                    'invalidSubmissionValue',
                    SlateJS::parseHTML( '<p>' . sprintf(
                    /* translators: %1$s is element Id, handled client-side */
                    __( 'Invalid value for element <mentions character="%1$s"></mentions>.', 'wp-eform' ),
                    $submission_element['id']
                ) . '</p>' . '<p>' . $validation_status['message'] . '</p>' )->getEditorChildrenJSON(),
                    null,
                    null
                ) );
            }
            
            // Form element okay and validated, now calculate score and proceed
            // with normalization and storage.
            // we need to pass the exact config and data the Element expects
            // although we store it in the form $config['type'], we can not pass it
            // that way, because get_data_score expects it like this
            $score_data = $element->get_data_score( $element_config, $submission_value );
            // add and get to the score, if primary score Id is present
            
            if ( $primary_score_id !== null && isset( $score_data[$primary_score_id] ) && isset( $this->score_data[$primary_score_id] ) ) {
                $primary_score = $score_data[$primary_score_id];
                
                if ( $primary_score['has'] ) {
                    $this->has_score = true;
                    $this->score_obtained += $primary_score['obtained'];
                    $this->max_score += $primary_score['outof'];
                }
            
            }
            
            // Add it to the cumulative score
            foreach ( $score_data as $score_id => $score_val ) {
                
                if ( isset( $this->score_data[$score_id] ) && $score_val['has'] ) {
                    $this->score_data[$score_id]['has'] = true;
                    $this->score_data[$score_id]['obtained'] += $score_val['obtained'];
                    $this->score_data[$score_id]['outof'] += $score_val['outof'];
                }
            
            }
            // add the already normalized value to the processed element
            $processed_element['value'][$form_element['type']] = $submission_value;
            // we have to transform the dictionary type of $score_data to indexed
            // array
            $processed_element['scores'] = convert_dictionary_to_list( $score_data, 'scoreId' );
            $processed_elements[] = $processed_element;
        }
        // Now check if some required element isn't present in the query
        // extract all of the ids from processed elements
        $processed_ids = extract_ids( $processed_elements );
        // now loop over all possible elements we have already extracted
        // and make sure that, if this is in the processed ids
        foreach ( $active_elements as $active_element_id ) {
            if ( !\in_array( $active_element_id, $processed_ids ) ) {
                throw new SubmissionValidationException( SubmissionValidationException::getSlateJSON(
                    'elementNotFound',
                    SlateJS::parseHTML( '<p>' . sprintf(
                    /* translators: %1$s is element Id, handled client-side */
                    __( 'Element <mentions character="%1$s"></mentions> is not found in data.', 'wp-eform' ),
                    $submission_element['id']
                ) . '</p>' )->getEditorChildrenJSON(),
                    null,
                    null
                ) );
            }
        }
        // All set, now override the internal data
        $this->processed_elements_data = $processed_elements;
        $this->processed_elements_dictionary = convert_list_to_dictionary( $processed_elements );
    }
    
    /**
     * Run mutations from settings config registry. Call `run_mutation` method on
     * every registered `ISettings`. This should be fired before saving the data
     * on model.
     *
     * @throws LogicException If data hasn't been processed yet.
     *
     * @param bool $is_update Whether or not this an update.
     * @param bool $admin_update       Whether or not this is an admin update.
     *
     * @return void
     */
    public function run_settings_mutations( bool $is_update, bool $admin_update ) : void
    {
        $this->throw_if_elements_not_processed();
        $registered_settings = Settings::get_registered_configs();
        $form_data = $this->get_form_data();
        $form_settings = $form_data['settings'];
        // Loop over all registered Settings config
        /** @var ISettings $s_instance */
        foreach ( $registered_settings as $s_key => $s_instance ) {
            // call run side effect on each and every one of them
            $s_instance->run_mutation(
                $this,
                $form_settings[$s_key],
                $is_update,
                $admin_update
            );
        }
    }
    
    /**
     * Process integrations for a submission. It expects the elements data to be
     * processed.
     *
     * @throws LogicException If data hasn't been processed yet.
     *
     * @param bool $is_update Whether or not this is an update operation.
     * @return void
     */
    public function run_integrations( bool $is_update ) : void
    {
        $this->throw_if_elements_not_processed();
        $registered_integrations = Integrations::get_registered_configs();
        $form_data = $this->form_model->read();
        $form_integrations = $form_data['integrations'];
        // Loop through all integrations and
        /** @var IIntegrations $i_instance */
        foreach ( $registered_integrations as $i_key => $i_instance ) {
            // Call integration with the element data passed
            $i_instance->do_integration( $this, $form_integrations[$i_key], $is_update );
        }
    }
    
    /**
     * Process payment intent and return data from the Payments class.
     * Call this when saving the form data and about to show the UI to the user.
     *
     * @throws LogicException If data hasn't been processed yet.
     *
     * @return mixed
     */
    public function process_payment_intent()
    {
        // 3. Loop through all payments and
        // 3a. Check if payment is a gateway and that is active in `general`.
        // 3b. Call the first process method and insert it to the response.
    }
    
    /**
     * Process all settings by calling each of the registered settings with
     * elements and their side effects.
     *
     * @throws LogicException If data hasn't been processed yet.
     *
     * @param bool $is_update Whether or not this is an update.
     * @param bool $admin_update Whether the update is done by an admin.
     *
     * @return void
     */
    public function run_settings_side_effects( bool $is_update, bool $admin_update ) : void
    {
        $this->throw_if_elements_not_processed();
        $registered_settings = Settings::get_registered_configs();
        $form_data = $this->get_form_data();
        $form_settings = $form_data['settings'];
        // Loop over all registered Settings config
        /** @var ISettings $s_instance */
        foreach ( $registered_settings as $s_key => $s_instance ) {
            // call run side effect on each and every one of them
            $s_instance->run_side_effect(
                $this,
                $form_settings[$s_key],
                $is_update,
                $admin_update
            );
        }
    }
    
    /**
     * Run Permissions Mutations. This should be run before things are saved
     * in the database.
     *
     * @param bool $is_update Whether or not this is for update.
     *
     * @return void
     */
    public function run_permissions_presave_mutations( bool $is_update ) : void
    {
        $this->throw_if_elements_not_processed();
        $registered_permissions = Permissions::get_registered_configs();
        $form_data = $this->get_form_data();
        $form_permissions = $form_data['permissions'];
        /** @var IPermissions $p_instance */
        foreach ( $registered_permissions as $p_key => $p_instance ) {
            $p_instance->run_presave_mutations( $is_update, $form_permissions[$p_key], $this );
        }
    }
    
    /**
     * Run Permissions Side effects. This should be run after things are saved
     * in the database.
     *
     * @param bool $is_update Whether or not this is for update.
     *
     * @return void
     */
    public function run_permissions_side_effects( bool $is_update ) : void
    {
        $this->throw_if_elements_not_processed();
        $registered_permissions = Permissions::get_registered_configs();
        $form_data = $this->get_form_data();
        $form_permissions = $form_data['permissions'];
        /** @var IPermissions $p_instance */
        foreach ( $registered_permissions as $p_key => $p_instance ) {
            $p_instance->run_postsave_side_effect(
                $is_update,
                $form_permissions[$p_key],
                $this->get_form_model(),
                new FormAuth(),
                $this
            );
        }
    }
    
    /**
     * Append data to Submission model to complete the Submission Type on GraphQL.
     *
     * @param array $data Original data as received from Submission model.
     * @param bool  $is_update Whether or not this is an update.
     *
     * @return array
     */
    public function append_data_to_model( array $data, bool $is_update ) : array
    {
        // TODO: Append some additional data needed during create and update operation
        // NOTE: Basically needed for payment related stuff.
        $data['paid'] = $this->get_paid_for_graphql();
        return $data;
    }
    
    /**
     * Finalize payemnt from the data received from Payments UI shown after
     * the intent was initiated. It calls Payment class with user input and
     * returns from the method.
     *
     * @throws LogicException If data hasn't been processed yet.
     *
     * @return mixed
     */
    public function finalize_payment()
    {
        $this->throw_if_elements_not_processed();
    }

}